# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2024, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""A type based dependency injection framework for Python 3.9+."""
from __future__ import annotations

__all__: list[str] = [
    "AllukaError",
    "AsyncOnlyError",
    "AsyncSelfInjecting",
    "BasicContext",
    "CachingContext",
    "Client",
    "Context",
    "Injected",
    "InjectedDescriptor",
    "MissingDependencyError",
    "OverridingContext",
    "SelfInjecting",
    "SyncOnlyError",
    "abc",
    "inject",
    "local",
]

from . import abc
from . import local
from ._client import Client
from ._client import inject
from ._context import BasicContext  # pyright: ignore[reportDeprecated]
from ._context import CachingContext
from ._context import Context
from ._context import OverridingContext
from ._errors import AllukaError
from ._errors import MissingDependencyError
from ._errors import SyncOnlyError
from ._self_injecting import AsyncSelfInjecting  # pyright: ignore[reportDeprecated]
from ._self_injecting import SelfInjecting  # pyright: ignore[reportDeprecated]
from ._types import Injected
from ._types import InjectedDescriptor

AsyncOnlyError = SyncOnlyError
